from __future__ import print_function

import argparse
import asyncio
import logging
import sys

from kinto_http import AsyncClient, cli_utils
from ruamel.yaml import YAML

from .kinto2yaml import introspect_server
from .logger import logger
from .validate import validate_export
from .yaml2kinto import initialize_server


async def execute():
    parser = argparse.ArgumentParser(description="Wizard to setup Kinto with YAML")
    subparsers = parser.add_subparsers(
        title="subcommand",
        description="Load/Dump/Validate",
        dest="subcommand",
        help="Choose and run with --help",
    )
    subparsers.required = True

    # load sub-command.
    subparser = subparsers.add_parser("load")
    subparser.set_defaults(which="load")
    cli_utils.add_parser_options(subparser)
    subparser.add_argument(dest="filepath", help="YAML file")
    subparser.add_argument(
        "--force",
        help="Load the file using the CLIENT_WINS conflict resolution strategy",
        action="store_true",
    )
    subparser.add_argument(
        "--dry-run", help="Do not apply write call to the server", action="store_true"
    )
    subparser.add_argument(
        "--delete-records", help="Delete records that are not in the file.", action="store_true"
    )
    subparser.add_argument(
        "--attachments", help="Load attachments from specified folder", default=None
    )
    subparser.add_argument(
        "--full",
        help="Load everything (same as with all --load-... options)",
        action="store_true",
        default=None,
    )
    for resource in ("bucket", "collection", "group", "record"):
        subparser.add_argument(
            f"--{resource}s",
            help=f"Load {resource}s",
            action="store_true",
            dest=f"load_{resource}s",
            default=None,
        )
    subparser.add_argument(
        "--data",
        help="Load attributes",
        action="store_true",
        dest="load_data",
        default=None,
    )
    subparser.add_argument(
        "--permissions",
        help="Load permissions",
        action="store_true",
        dest="load_permissions",
        default=None,
    )

    # dump sub-command.
    subparser = subparsers.add_parser("dump")
    subparser.set_defaults(which="dump")
    cli_utils.add_parser_options(subparser)
    subparser.add_argument(
        "--full",
        help="Full output (same as with --data, --permissions, --collections, --groups and --records options)",
        action="store_true",
    )
    for resource in ("bucket", "collection", "group", "record"):
        subparser.add_argument(
            f"--{resource}s",
            help=f"Export {resource}s",
            action="store_true",
            dest=f"dump_{resource}s",
            default=None,
        )
    subparser.add_argument(
        "--data",
        help="Export buckets, collections and groups data",
        action="store_true",
        default=False,
    )
    subparser.add_argument(
        "--permissions",
        help="Export buckets, collections and groups permissions",
        action="store_true",
        default=False,
    )
    subparser.add_argument(
        "--attachments", help="Export collections' attachments to specified folder", default=None
    )

    # validate sub-command.
    subparser = subparsers.add_parser("validate")
    subparser.set_defaults(which="validate")
    subparser.set_defaults(verbosity=logging.INFO)
    subparser.add_argument(dest="filepath", help="YAML file to validate")
    cli_utils.add_parser_options(subparser)

    # Parse CLI args.
    args = parser.parse_args()
    cli_utils.setup_logger(logger, args)
    kinto_logger = logging.getLogger("kinto_http")
    cli_utils.setup_logger(kinto_logger, args)

    if args.which == "validate":
        logger.debug("Start validation...")
        logger.info("Load YAML file {!r}".format(args.filepath))
        yaml = YAML(typ="safe")
        with open(args.filepath, "r") as f:
            config = yaml.load(f)
        logger.info("File loaded!")
        fine = validate_export(config)
        sys.exit(0 if fine else 1)

    logger.debug("Instantiate Kinto client.")
    # TODO: add cli_utils.create_async_client_from_args(args)
    async_client = AsyncClient(
        server_url=args.server,
        auth=args.auth,
        bucket=getattr(args, "bucket", None),
        collection=getattr(args, "collection", None),
        retry=args.retry,
        retry_after=args.retry_after,
        dry_mode=getattr(args, "dry_run", False),
        ignore_batch_4xx=args.ignore_batch_4xx,
    )

    # Run chosen subcommand.
    if args.which == "dump":
        if args.full:
            records = True
            buckets = True
            collections = True
            groups = True
            attachments = args.attachments or "__attachments__"
            data = True
            permissions = True
        else:
            records = args.dump_records
            buckets = args.dump_buckets
            collections = args.dump_collections
            groups = args.dump_groups
            attachments = args.attachments
            data = args.data
            permissions = args.permissions

        logger.debug(
            "Start introspection with %s...",
            " and ".join(
                filter(
                    None,
                    [
                        "data" if data else None,
                        "permissions" if permissions else None,
                        "collections" if collections else None,
                        "groups" if groups else None,
                        "records" if records else None,
                        "attachments" if attachments else None,
                    ],
                )
            ),
        )

        result = await introspect_server(
            async_client,
            bucket=args.bucket,
            collection=args.collection,
            data=data,
            permissions=permissions,
            buckets=buckets,
            collections=collections,
            groups=groups,
            records=records,
            attachments=attachments,
        )
        yaml = YAML()
        yaml.default_flow_style = False
        yaml.dump(result, sys.stdout)

    elif args.which == "load":
        # If --full is passed or not any --records, etc. specified
        if args.full or not any(
            (args.load_buckets, args.load_collections, args.load_records, args.load_groups)
        ):
            load_buckets = True
            load_collections = True
            load_records = True
            load_groups = True
        else:
            load_buckets = args.load_buckets
            load_collections = args.load_collections
            load_records = args.load_records
            load_groups = args.load_groups
        # If --full is passed or --data and --permissions not specified
        if args.full or (not args.load_data and not args.load_permissions):
            load_data = True
            load_permissions = True
        else:
            load_data = args.load_data
            load_permissions = args.load_permissions

        logger.debug("Start initialization...")
        logger.info("Load YAML file {!r}".format(args.filepath))
        yaml = YAML(typ="safe")
        with open(args.filepath, "r") as f:
            config = yaml.load(f)
        await initialize_server(
            async_client,
            config,
            bucket=args.bucket,
            collection=args.collection,
            force=args.force,
            delete_missing_records=args.delete_records,
            attachments=args.attachments,
            load_buckets=load_buckets,
            load_collections=load_collections,
            load_records=load_records,
            load_groups=load_groups,
            load_data=load_data,
            load_permissions=load_permissions,
        )


def main():
    asyncio.run(execute())



import asyncio
import itertools
import os

from kinto_http import exceptions as kinto_exceptions

from .logger import logger


MAX_PARALLEL_REQUESTS = 8


def sorted_principals(permissions):
    return {perm: sorted(principals) for perm, principals in sorted(permissions.items())}


async def gather_dict(dct):
    """Utility function to do asyncio.gather for dictionaries

    Do asyncio.gather on the dictionary's values, and return the
    dictionary with the futures' results replacing the futures.

    >>> promise = Future()
    >>> promise.set_result("value")
    >>> await gather_dict({"key": promise})
    {"key": "value"}
    """
    items = dct.items()
    results = await asyncio.gather(*(item[1] for item in items))
    keys_results = zip((item[0] for item in items), results)
    return {k: v for k, v in keys_results if v is not None}


async def introspect_server(
    client,
    bucket=None,
    collection=None,
    data=False,
    permissions=True,
    buckets=True,
    collections=True,
    groups=True,
    records=False,
    attachments=None,
):
    if bucket:
        logger.info("Only inspect bucket `{}`.".format(bucket))
        bucket_info = await introspect_bucket(
            client,
            bucket,
            collection=collection,
            data=data,
            permissions=permissions,
            buckets=buckets,
            collections=collections,
            groups=groups,
            records=records,
            attachments=attachments,
        )
        if bucket_info:
            return {"buckets": {bucket: bucket_info}}
        return {"buckets": {}}

    logger.info("Fetch buckets list.")
    buckets = await client.get_buckets()
    buckets_tree = await gather_dict(
        {
            bucket["id"]: introspect_bucket(
                client,
                bucket["id"],
                collection=collection,
                data=data,
                permissions=permissions,
                buckets=buckets,
                collections=collections,
                groups=groups,
                records=records,
                attachments=attachments,
            )
            for bucket in buckets
        }
    )
    return {"buckets": buckets_tree}


async def introspect_bucket(
    client,
    bid,
    collection=None,
    data=False,
    permissions=True,
    buckets=True,
    collections=True,
    groups=True,
    records=False,
    attachments=None,
):
    logger.info("Fetch information of bucket {!r}".format(bid))
    try:
        bucket = await client.get_bucket(id=bid)
    except kinto_exceptions.BucketNotFound:
        logger.error("Could not read bucket {!r}".format(bid))
        return None

    if collection:
        try:
            result = {
                "collections": {
                    collection: await introspect_collection(
                        client,
                        bid,
                        collection,
                        data=data,
                        permissions=permissions,
                        collections=collections,
                        records=records,
                        attachments=attachments,
                    )
                },
            }
        except kinto_exceptions.CollectionNotFound:
            return None
    else:
        result = {}
        if collections:
            result["collections"] = await gather_dict(
                {
                    collection["id"]: introspect_collection(
                        client,
                        bid,
                        collection["id"],
                        data=data,
                        permissions=permissions,
                        collections=collections,
                        records=records,
                        attachments=attachments,
                    )
                    for collection in (await client.get_collections(bucket=bid))
                }
            )

        if groups:
            result["groups"] = await gather_dict(
                {
                    group["id"]: introspect_group(
                        client, bid, group["id"], data=data, permissions=permissions
                    )
                    for group in (await client.get_groups(bucket=bid))
                }
            )

    if buckets and permissions:
        if len(bucket["permissions"]) == 0:
            logger.warning(
                "⚠️ Could not read permissions of bucket {!r}".format(bid)
            )  # pragma: no cover
        result["permissions"] = sorted_principals(bucket["permissions"])

    if buckets and data:
        result["data"] = bucket["data"]

    return result


async def introspect_collection(
    client,
    bid,
    cid,
    data=False,
    permissions=True,
    collections=True,
    records=False,
    attachments=None,
):
    logger.info("Fetch information of collection {!r}/{!r}".format(bid, cid))
    collection = await client.get_collection(bucket=bid, id=cid)

    result = {}

    if collections and permissions:
        if len(collection["permissions"]) == 0:
            logger.warning(
                "⚠️ Could not read permissions of collection {!r}/{!r}".format(bid, cid)
            )  # pragma: no cover
        result["permissions"] = sorted_principals(collection["permissions"])

    if collections and data:
        result["data"] = collection["data"]

    if records or attachments:
        records = await client.get_records(bucket=bid, collection=cid)
        result["records"] = {
            # XXX: we don't show permissions, until we have a way to fetch records
            # in batch (see Kinto/kinto-http.py#145)
            record["id"]: {"data": record, "permissions": {}} if permissions else {"data": record}
            for record in records
        }

    if attachments:
        futures = [
            client.download_attachment(
                record,
                filepath=os.path.join(attachments, record["attachment"]["location"]),
                save_metadata=True,
            )
            for record in records
            if "attachment" in record
        ]
        if futures:
            logger.info("Dump attachments to %s", attachments)
        chunks = itertools.batched(futures, MAX_PARALLEL_REQUESTS)
        for chunk in chunks:
            await asyncio.gather(*chunk)

    return result


async def introspect_group(client, bid, gid, data=False, permissions=True):
    logger.info("Fetch information of group {!r}/{!r}".format(bid, gid))
    group = await client.get_group(bucket=bid, id=gid)

    result = {}

    if permissions:
        if len(group["permissions"]) == 0:
            logger.warning(
                "⚠️ Could not read permissions of group {!r}/{!r}".format(bid, gid)
            )  # pragma: no cover
        result["permissions"] = sorted_principals(group["permissions"])

    data = group["data"] if data else {}
    data["members"] = group["data"]["members"]
    result["data"] = data

    return result



import builtins
import io
import json
import os
import shutil
import sys
import unittest
from contextlib import contextmanager, redirect_stdout
from copy import deepcopy

import pytest
import requests
from kinto_http import Client, exceptions
from ruamel.yaml import YAML

from kinto_wizard.__main__ import main


def load(server, auth, file, bucket=None, collection=None, extra=None):
    cmd = "kinto-wizard {} --server={} --auth={}"

    if bucket:
        cmd += " --bucket={}".format(bucket)

    if collection:
        cmd += " --collection={}".format(collection)

    if extra:
        cmd += " " + extra

    load_cmd = cmd.format("load {}".format(file), server, auth)
    sys.argv = load_cmd.strip().split(" ")
    return main()


def dump(server, auth, bucket=None, collection=None, extra=None):
    cmd = "kinto-wizard {} --server={} --auth={}"
    dump_cmd = cmd.format("dump --full", server, auth)

    if bucket:
        dump_cmd += " --bucket={}".format(bucket)

    if collection:
        dump_cmd += " --collection={}".format(collection)

    if extra:
        dump_cmd += " " + extra

    sys.argv = dump_cmd.split(" ")
    output = io.StringIO()
    with redirect_stdout(output):
        main()
    output.flush()

    # Check that identical to original file.
    return output.getvalue()


def validate(filename):
    sys.argv = ["kinto-wizard", "validate", filename]
    return main()


def assert_identical(a, b):
    yaml = YAML(typ="safe")
    a_parsed = yaml.load(a)
    b_parsed = yaml.load(b)
    assert a_parsed == b_parsed


class FunctionalTest(unittest.TestCase):
    server = os.getenv("SERVER_URL", "http://localhost:8888/v1")
    auth = os.getenv("AUTH", "user:pass")
    file = os.getenv("FILE", "tests/kinto.yaml")

    def setUp(self):
        requests.post(self.server + "/__flush__")

    def load(self, bucket=None, collection=None, filename=None, extra=None):
        return load(self.server, self.auth, filename or self.file, bucket, collection, extra)

    def dump(self, bucket=None, collection=None, extra=None):
        return dump(self.server, self.auth, bucket, collection, extra)

    def validate(self, filename=None, code=0):
        try:
            validate(filename or self.file)
        except SystemExit as e:
            if e.code == code:
                return
            else:
                self.fail(f"Unexpected validation status {e.code} != {code}")


class DryRunLoad(FunctionalTest):
    def test_dry_round_trip(self):
        cmd = "kinto-wizard {} --server={} --auth={} --dry-run"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()
        client = Client(server_url=self.server, auth=tuple(self.auth.split(":")))
        with pytest.raises(exceptions.KintoException):
            client.get_bucket(id="staging")


@contextmanager
def mockInput(mock):
    original_input = builtins.input
    builtins.input = lambda _: mock
    yield
    builtins.input = original_input


class SimpleDump(FunctionalTest):
    def test_round_trip(self):
        cmd = "kinto-wizard {} --server={} --auth={}"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()

        dump_cmd = (
            cmd.format("dump", self.server, self.auth) + " --permissions --collections --groups"
        )
        sys.argv = dump_cmd.split(" ")
        output = io.StringIO()
        with redirect_stdout(output):
            main()
        output.flush()

        # Check that identical to original file.
        generated = output.getvalue()
        with open(self.file) as f:
            assert_identical(f.read(), generated)


class FullDump(FunctionalTest):
    file = os.getenv("FILE", "tests/kinto-full.yaml")

    def test_round_trip(self):
        # Load some data
        cmd = "kinto-wizard {} --server={} --auth={}"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()

        cmd = "kinto-wizard {} --server={} --auth={} --full"
        load_cmd = cmd.format("dump", self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        output = io.StringIO()
        with redirect_stdout(output):
            main()
        output.flush()

        # Check that identical to original file.
        generated = output.getvalue()
        with open(self.file) as f:
            assert_identical(f.read(), generated)

    def test_round_trip_with_client_wins(self):
        # Load some data
        cmd = "kinto-wizard {} --server={} --auth={}"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()

        # Change something that could make the server to fail.
        client = Client(server_url=self.server, auth=tuple(self.auth.split(":")))
        client.update_record(
            bucket="build-hub",
            collection="archives",
            id="0831d549-0a69-48dd-b240-feef94688d47",
            data={},
        )
        record = client.get_record(
            bucket="build-hub", collection="archives", id="0831d549-0a69-48dd-b240-feef94688d47"
        )
        assert set(record["data"].keys()) == {"id", "last_modified"}
        cmd = "kinto-wizard {} --server={} -D --auth={} --force"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()
        record = client.get_record(
            bucket="build-hub", collection="archives", id="0831d549-0a69-48dd-b240-feef94688d47"
        )
        assert set(record["data"].keys()) != {"id", "last_modified"}

    def test_round_trip_with_client_wins_and_delete_missing_records(self):
        # Load some data
        cmd = "kinto-wizard {} --server={} --auth={}"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()

        # Change something that could make the server to fail.
        client = Client(server_url=self.server, auth=tuple(self.auth.split(":")))
        client.create_record(
            bucket="build-hub",
            collection="archives",
            id="8031d549-0a69-48dd-b240-feef94688d47",
            data={},
        )
        cmd = "kinto-wizard {} --server={} -D --auth={} --force --delete-records"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()
        with pytest.raises(exceptions.KintoException) as exc:
            client.get_record(
                bucket="build-hub",
                collection="archives",
                id="8031d549-0a69-48dd-b240-feef94688d47",
            )
        assert "'Not Found'" in str(exc.value)

    def test_round_trip_with_delete_missing_records_ask_for_confirmation(self):
        # Load some data
        cmd = "kinto-wizard {} --server={} --auth={}"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()

        # Change something that could make the server to fail.
        client = Client(server_url=self.server, auth=tuple(self.auth.split(":")))
        client.create_record(
            bucket="build-hub",
            collection="archives",
            id="8031d549-0a69-48dd-b240-feef94688d47",
            data={},
        )
        cmd = "kinto-wizard {} --server={} -D --auth={} --delete-records"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")

        with mockInput("yes"):
            main()

        with pytest.raises(exceptions.KintoException) as exc:
            client.get_record(
                bucket="build-hub",
                collection="archives",
                id="8031d549-0a69-48dd-b240-feef94688d47",
            )
        assert "'Not Found'" in str(exc.value)

    def test_round_trip_with_delete_missing_records_handle_misconfirmation(self):
        # Load some data
        cmd = "kinto-wizard {} --server={} --auth={}"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()

        # Change something that could make the server to fail.
        client = Client(server_url=self.server, auth=tuple(self.auth.split(":")))
        client.create_record(
            bucket="build-hub",
            collection="archives",
            id="8031d549-0a69-48dd-b240-feef94688d47",
            data={},
        )
        cmd = "kinto-wizard {} --server={} -D --auth={} --delete-records"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")

        with mockInput("no"):
            with pytest.raises(SystemExit):
                main()


class DataRecordsDump(FunctionalTest):
    file = os.getenv("FILE", "tests/kinto-full.yaml")

    def test_round_trip(self):
        # Load some data
        cmd = "kinto-wizard {} --server={} --auth={}"
        load_cmd = cmd.format("load {}".format(self.file), self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        main()

        cmd = "kinto-wizard {} --server={} --auth={} --data --collections --groups --records --permissions"
        load_cmd = cmd.format("dump", self.server, self.auth)
        sys.argv = load_cmd.split(" ")
        output = io.StringIO()
        with redirect_stdout(output):
            main()
        output.flush()

        # Check that identical to original file.
        generated = output.getvalue()
        with open(self.file) as f:
            assert_identical(f.read(), generated)


class BucketCollectionSelectionableDump(FunctionalTest):
    file = os.getenv("FILE", "tests/dumps/dump-full.yaml")

    def test_validate(self):
        self.validate()

    def test_round_trip_with_bucket_selection_on_load(self):
        self.load(bucket="natim")
        generated = self.dump()
        with open("tests/dumps/dump-natim.yaml") as f:
            assert_identical(f.read(), generated)

    def test_round_trip_with_bucket_selection(self):
        self.load()
        generated = self.dump(bucket="natim")
        with open("tests/dumps/dump-natim.yaml") as f:
            assert_identical(f.read(), generated)

    def test_round_trip_with_bucket_collection_selection_on_load(self):
        self.load(bucket="natim", collection="toto")
        generated = self.dump()
        with open("tests/dumps/dump-natim-toto-groups.yaml") as f:
            assert_identical(f.read(), generated)

    def test_round_trip_with_bucket_collection_selection(self):
        self.load()
        generated = self.dump(bucket="natim", collection="toto")
        with open("tests/dumps/dump-natim-toto.yaml") as f:
            assert_identical(f.read(), generated)

    def test_round_trip_with_collection_selection_on_load(self):
        self.load(collection="toto")
        generated = self.dump()
        with open("tests/dumps/dump-toto-groups.yaml") as f:
            assert_identical(f.read(), generated)

    def test_round_trip_with_collection_selection(self):
        self.load()
        generated = self.dump(collection="toto")
        with open("tests/dumps/dump-toto.yaml") as f:
            assert_identical(f.read(), generated)

    def test_wizard_can_handle_dates(self):
        self.load(bucket="date")
        generated = self.dump()
        with open("tests/dumps/dump-date.yaml") as f:
            assert_identical(f.read(), generated)


class YAMLReferenceSupportTest(FunctionalTest):
    file = os.getenv("FILE", "tests/dumps/with-references.yaml")

    def test_validate(self):
        self.validate()

    def test_file_can_have_yaml_references(self):
        self.load()

        client = Client(server_url=self.server, auth=tuple(self.auth.split(":")))

        collection = client.get_collection(bucket="main", id="certificates")
        assert "url" in collection["data"]["schema"]["properties"]
        collection = client.get_collection(bucket="main", id="addons")
        assert "url" in collection["data"]["schema"]["properties"]

        # the anchor did not get interpreted as a bucket:
        with self.assertRaises(exceptions.KintoException):
            client.get_collection(bucket="attachment-schema")


class WrongSchemaValidationTest(FunctionalTest):
    file = "tests/dumps/wrong-schema.yaml"

    def test_validate(self):
        self.validate(code=1)


class MiscUpdates(FunctionalTest):
    def get_client(self):
        return Client(server_url=self.server, auth=tuple(self.auth.split(":")))

    def test_validate(self):
        # This dump has a schema that requires `title` field, and a record doesn't have it.
        self.validate(filename="tests/dumps/with-schema-1.yaml", code=1)
        # This dump has a schema that does not require `title` field, so the dump is valid.
        self.validate(filename="tests/dumps/with-schema-2.yaml")

    def test_raises_with_4xx_error_in_batch(self):
        with pytest.raises(exceptions.KintoBatchException):
            self.load(filename="tests/dumps/with-schema-1.yaml")
        records = self.get_client().get_records(bucket="natim", collection="toto")
        assert len(records) == 0

    def test_ignore_batch_4xx_errors_if_specified(self):
        # Raises a KintoBatchException in case of error
        self.load(filename="tests/dumps/with-schema-1.yaml", extra="--ignore-batch-4xx")

    def test_record_updates(self):
        self.load(filename="tests/dumps/with-schema-1.yaml", extra="--ignore-batch-4xx")
        client = self.get_client()
        client.create_record(
            data={"title": "titi", "last_modified": 1496132479110},
            id="e2686bac-c45e-4144-9738-edfeb3d9da6d",
            collection="toto",
            bucket="natim",
        )
        self.load(filename="tests/dumps/with-schema-2.yaml")
        r = client.get_record(
            id="e2686bac-c45e-4144-9738-edfeb3d9da6d", collection="toto", bucket="natim"
        )
        assert r["data"]["title"] == "toto"

    def test_group_updates(self):
        self.load(filename="tests/dumps/with-groups.yaml")
        client = self.get_client()
        client.update_group(
            data={"members": ["alexis", "mathieu", "remy"]}, id="toto", bucket="natim"
        )
        self.load(filename="tests/dumps/with-groups.yaml")
        r = client.get_group(id="toto", bucket="natim")
        assert r["data"]["members"] == ["mathieu", "alexis"]

    def test_timestamps_are_not_changed_if_dump_hasnt_changed(self):
        self.load(filename="tests/dumps/with-groups.yaml")
        client = self.get_client()
        before = client.get_group(id="toto", bucket="natim")["data"]["last_modified"]

        self.load(filename="tests/dumps/with-groups.yaml")

        after = client.get_group(id="toto", bucket="natim")["data"]["last_modified"]
        assert before == after


class AttachmentsTest(FunctionalTest):
    def setUp(self):
        os.makedirs("/tmp/__attachments__/main/archives/", exist_ok=True)
        return super().setUp()

    def tearDown(self):
        shutil.rmtree("/tmp/__attachments__", ignore_errors=True)
        return super().tearDown()

    @property
    def client(self):
        return Client(server_url=self.server, auth=tuple(self.auth.split(":")))

    def _create_attachment_manually(self):
        self.client.create_bucket(id="main")
        self.client.create_collection(bucket="main", id="archives")
        return self.client.add_attachment(
            id="abc",
            bucket="main",
            collection="archives",
            filepath="tests/dumps/image.jpg",
            data={"title": "Test Attachment"},
        )

    def test_dump_with_attachments(self):
        self._create_attachment_manually()

        dumped = self.dump(extra="--attachments=/tmp/__attachments__")
        yaml = YAML(typ="safe")
        dumped_parsed = yaml.load(dumped)

        with open("tests/dumps/with-attachments.yaml") as f:
            expected = f.read()
        expected_parsed = yaml.load(expected)

        # The location is randomly assigned by server.
        real_location = dumped_parsed["buckets"]["main"]["collections"]["archives"]["records"][
            "abc"
        ]["data"]["attachment"]["location"]

        expected_parsed["buckets"]["main"]["collections"]["archives"]["records"]["abc"]["data"][
            "attachment"
        ]["location"] = real_location
        dumped_without_last_modified = deepcopy(dumped_parsed)
        del dumped_without_last_modified["buckets"]["main"]["data"]["last_modified"]
        del dumped_without_last_modified["buckets"]["main"]["collections"]["archives"]["data"][
            "last_modified"
        ]
        del dumped_without_last_modified["buckets"]["main"]["collections"]["archives"]["records"][
            "abc"
        ]["data"]["last_modified"]

        assert dumped_without_last_modified == expected_parsed
        assert os.path.exists("/tmp/__attachments__")
        assert os.path.exists(os.path.join("/tmp/__attachments__", real_location))

    def test_load_with_attachments_from_unexisting_folder(self):
        self._create_attachment_manually()

        self.load(filename="tests/dumps/with-attachments.yaml", extra="--attachments=/tmp/missing")

    def test_load_with_attachments_from_unexisting_folder_with_existing_record(self):
        self.load(filename="tests/dumps/with-attachments.yaml", extra="--attachments=/tmp/missing")

    def test_load_with_attachments_with_missing_record(self):
        shutil.copyfile(
            "tests/dumps/image.jpg",
            "/tmp/__attachments__/main/archives/aedddd6b-f6ef-423b-8e4c-ac23a74736c3.jpg",
        )

        self.load(
            filename="tests/dumps/with-attachments.yaml",
            extra="--attachments=/tmp/__attachments__",
        )

        record_after = self.client.get_record(bucket="main", collection="archives", id="abc")
        assert (
            record_after["data"]["attachment"]["filename"]
            == "aedddd6b-f6ef-423b-8e4c-ac23a74736c3.jpg"
        )
        assert record_after["data"]["attachment"]["size"] > 0

    def test_load_with_attachments_uses_filename_from_meta_file(self):
        location = "/tmp/__attachments__/main/archives/aedddd6b-f6ef-423b-8e4c-ac23a74736c3.jpg"
        shutil.copyfile("tests/dumps/image.jpg", location)
        with open(location + ".meta.json", "w") as f:
            json.dump({"attachment": {"filename": "image.jpg"}}, f)

        self.load(
            filename="tests/dumps/with-attachments.yaml",
            extra="--attachments=/tmp/__attachments__",
        )

        record_after = self.client.get_record(bucket="main", collection="archives", id="abc")
        assert record_after["data"]["attachment"]["filename"] == "image.jpg"
        assert record_after["data"]["attachment"]["size"] > 0

    def test_load_with_attachments_with_existing_record(self):
        attachment_before = self._create_attachment_manually()
        shutil.copyfile(
            "tests/dumps/image.jpg",
            "/tmp/__attachments__/main/archives/aedddd6b-f6ef-423b-8e4c-ac23a74736c3.jpg",
        )

        self.load(
            filename="tests/dumps/with-attachments.yaml",
            extra="--attachments=/tmp/__attachments__",
        )

        record_after = self.client.get_record(bucket="main", collection="archives", id="abc")
        assert attachment_before["hash"] == record_after["data"]["attachment"]["hash"]

    def test_load_with_attachments_with_existing_record_but_different(self):
        attachment_before = self._create_attachment_manually()
        with open(
            "/tmp/__attachments__/main/archives/aedddd6b-f6ef-423b-8e4c-ac23a74736c3.jpg", "wb"
        ) as f:
            f.write(b"Different content")

        self.load(
            filename="tests/dumps/with-attachments.yaml",
            extra="--attachments=/tmp/__attachments__",
        )

        record_after = self.client.get_record(bucket="main", collection="archives", id="abc")
        assert attachment_before["hash"] != record_after["data"]["attachment"]["hash"]


class PartialLoadTest(FunctionalTest):
    def test_load_only_groups(self):
        self.load(filename="tests/dumps/dump-full.yaml", extra="--groups")

    def test_load_only_permissions(self):
        self.load(filename="tests/dumps/dump-full.yaml", extra="--permissions")



try:
    from jsonschema import Draft7Validator as SchemaValidator
except ImportError:  # pragma: no cover
    from jsonschema import Draft4Validator as SchemaValidator
from jsonschema import SchemaError, ValidationError, validate

from .logger import logger


IGNORED_FIELDS = (
    "id",
    "last_modified",
    "schema",
    "attachment",
)


def check_schema(data):
    try:
        SchemaValidator.check_schema(data)
    except SchemaError as e:
        message = e.path.pop() + e.message
        raise ValidationError(message)


def validate_schema(data, schema, ignore_fields=[]):
    required_fields = [f for f in schema.get("required", []) if f not in ignore_fields]
    # jsonschema doesn't accept 'required': [] yet.
    # See https://github.com/Julian/jsonschema/issues/337.
    # In the meantime, strip out 'required' if no other fields are required.
    if required_fields:
        schema = {**schema, "required": required_fields}
    else:
        schema = {f: v for f, v in schema.items() if f != "required"}

    data = {f: v for f, v in data.items() if f not in ignore_fields}

    try:
        validate(data, schema)
    except ValidationError as e:  # pragma: no cover
        if e.path:
            field = e.path[-1]
        elif e.validator_value:
            field = e.validator_value[-1]
        else:
            field = e.schema_path[-1]
        e.field = field
        raise e


def validate_export(config):
    everything_is_fine = True
    if "buckets" in config:
        buckets = config.get("buckets", {})
    else:  # pragma: no cover
        # Legacy for file before kinto-wizard 4.0
        logger.warning(
            "Your file seems to be in legacy format. Please add a `buckets:` root level."
        )
        buckets = config
    for bid, bucket in buckets.items():
        logger.info(f"- Bucket {bid}")
        bucket_collections = bucket.get("collections", {})
        for cid, collection in bucket_collections.items():
            logger.info(f"  - Collection {cid}")
            collection_data = collection.get("data", {})
            if "schema" not in collection_data:
                logger.info("    No schema\n")
                continue

            schema = collection_data["schema"]
            try:
                check_schema(schema)
            except ValidationError:
                logger.exception(f"Collection {cid!r} validation failed.")
                everything_is_fine = False
                continue

            collection_records = collection.get("records", {})

            for record_id, record in collection_records.items():
                try:
                    validate_schema(record["data"], schema, ignore_fields=IGNORED_FIELDS)
                except ValidationError:
                    logger.exception(f"Record {record_id!r} validation failed.")
                    everything_is_fine = False
    return everything_is_fine



from __future__ import print_function

import json
import os

from .kinto2yaml import introspect_server, sorted_principals
from .logger import logger


async def initialize_server(
    async_client,
    config,
    bucket=None,
    collection=None,
    force=False,
    delete_missing_records=False,
    attachments=None,
    load_buckets=True,
    load_collections=True,
    load_records=True,
    load_groups=True,
    load_data=True,
    load_permissions=True,
):
    logger.debug("Converting YAML config into a server batch.")
    bid = bucket
    cid = collection
    # 1. Introspect current server state.
    if not force or delete_missing_records:
        current_server_status = await introspect_server(
            async_client, bucket=bucket, collection=collection, records=True
        )
        existing_server_buckets = current_server_status["buckets"]
    else:
        # We don't need to load it because we will override it nevertheless.
        existing_server_buckets = {}
    # 2. For each bucket
    buckets = config["buckets"]
    with await async_client.batch() as batch:
        for bucket_id, bucket in buckets.items():
            # Skip buckets that we don't want to import.
            if bid and bucket_id != bid:
                logger.debug("Skip bucket {}".format(bucket_id))
                continue
            bucket_exists = bucket_id in existing_server_buckets
            bucket_data = bucket.get("data", {}) if load_data else {}
            bucket_permissions = (
                sorted_principals(bucket.get("permissions", {})) if load_permissions else {}
            )
            bucket_groups = bucket.get("groups", {}) if load_groups else {}
            bucket_collections = bucket.get("collections", {}) if load_collections else {}

            # Skip bucket if we don't have a collection in them
            if cid and cid not in bucket_collections:
                logger.debug("Skip bucket {}".format(bucket_id))
                continue

            if load_buckets:
                if not bucket_exists:
                    existing_bucket_groups = {}
                    existing_bucket_collections = {}

                    # Create the bucket if not present in the introspection
                    await batch.create_bucket(
                        id=bucket_id,
                        data=bucket_data if load_data else None,
                        permissions=bucket_permissions if load_permissions else None,
                        safe=(not force),
                    )
                else:
                    existing_bucket = existing_server_buckets[bucket_id]
                    existing_bucket_groups = {}
                    existing_bucket_collections = {}
                    existing_bucket_data = {}
                    existing_bucket_permissions = {}

                    if existing_bucket:
                        existing_bucket_groups = existing_bucket.get("groups", {})
                        existing_bucket_collections = existing_bucket.get("collections", {})

                        # Patch the bucket if mandatory
                        existing_bucket_data = existing_bucket.get("data", {})
                        existing_bucket_permissions = existing_bucket.get("permissions", {})

                    if (
                        existing_bucket_data != bucket_data
                        or existing_bucket_permissions != bucket_permissions
                    ):
                        await batch.patch_bucket(
                            id=bucket_id,
                            data=bucket_data if load_data else None,
                            permissions=bucket_permissions if load_permissions else None,
                        )

            # 2.1 For each group, patch it if needed
            if load_groups:
                for group_id, group_info in bucket_groups.items():
                    group_exists = bucket_exists and group_id in existing_bucket_groups
                    group_data = group_info.get("data", {}) if load_data else {}
                    group_permissions = (
                        sorted_principals(group_info.get("permissions", {}))
                        if load_permissions
                        else {}
                    )

                    if not group_exists:
                        await batch.create_group(
                            id=group_id,
                            bucket=bucket_id,
                            data=group_data if load_data else None,
                            permissions=group_permissions if load_permissions else None,
                            safe=(not force),
                        )
                    else:
                        existing_group = existing_bucket_groups[group_id]
                        existing_group_data = existing_group.get("data", {})
                        existing_group_permissions = existing_group.get("permissions", {})

                        if (
                            existing_group_data != group_data
                            or existing_group_permissions != group_permissions
                        ):
                            await batch.patch_group(
                                id=group_id,
                                bucket=bucket_id,
                                data=group_data if load_data else None,
                                permissions=group_permissions if load_permissions else None,
                            )

            # 2.2 For each collection patch it if mandatory
            if load_collections:
                for collection_id, collection in bucket_collections.items():
                    # Skip collections that we don't want to import.
                    if cid and collection_id != cid:
                        logger.debug("Skip collection {}/{}".format(bucket_id, collection_id))
                        continue
                    collection_exists = (
                        bucket_exists and collection_id in existing_bucket_collections
                    )
                    collection_data = collection.get("data", {})
                    collection_permissions = sorted_principals(collection.get("permissions", {}))

                    if not collection_exists:
                        await batch.create_collection(
                            id=collection_id,
                            bucket=bucket_id,
                            data=collection_data if load_data else None,
                            permissions=collection_permissions if load_permissions else None,
                            safe=(not force),
                        )
                    else:
                        existing_collection = existing_bucket_collections[collection_id]
                        existing_collection_data = existing_collection.get("data", {})
                        existing_collection_permissions = existing_collection.get(
                            "permissions", {}
                        )

                        if (
                            existing_collection_data != collection_data
                            or existing_collection_permissions != collection_permissions
                        ):
                            await batch.patch_collection(
                                id=collection_id,
                                bucket=bucket_id,
                                data=collection_data if load_data else None,
                                permissions=collection_permissions if load_permissions else None,
                            )
        logger.debug("Sending batch:\n\n%s" % batch.session.requests)
    logger.info("Buckets, groups, and collections uploaded")

    if not load_records:
        # We're done here.
        return

    with await async_client.batch() as batch:
        for bucket_id, bucket in buckets.items():
            if bid and bucket_id != bid:
                continue

            bucket_collections = bucket.get("collections", {})
            existing_bucket = existing_server_buckets.get(bucket_id, {})
            existing_bucket_collections = existing_bucket.get("collections", {})

            for collection_id, collection in bucket_collections.items():
                if cid and collection_id != cid:
                    continue

                existing_collection = existing_bucket_collections.get(collection_id)
                existing_records = (
                    existing_collection.get("records", {}) if existing_collection else {}
                )
                collection_exists = existing_collection is not None

                # For each collection, create its records.
                collection_records = collection.get("records", {})
                for record_id, record in collection_records.items():
                    record_exists = collection_exists and record_id in existing_records
                    record_data = record.get("data", {})
                    record_permissions = sorted_principals(record.get("permissions", None))

                    # If 'attachment' field is present on record, then we look whether we have
                    # to upload it from the local folder `attachments`.
                    # If the collection has a JSON schema where the attachment field is mandatory,
                    # creation will fail.
                    # But if the attachment is not present, we warn and try anyway.
                    must_upload_attachment = False
                    if attachments is not None and "attachment" in record_data:
                        location = record_data["attachment"]["location"]
                        attachment_path = os.path.join(attachments, location)

                        if not os.path.exists(attachment_path):
                            # No local file, we simply ignore the 'attachment' field
                            # and will proceed with the upsert below.
                            record_data.pop("attachment")
                            if not record_exists:
                                # For creations, warn because it may fail because of mandatory field in JSON schema.
                                logger.warning(
                                    "Attachment for %s/%s/%s not found: %s",
                                    bucket_id,
                                    collection_id,
                                    record_id,
                                    attachment_path,
                                )
                        else:
                            # Attachment exists on disk.
                            must_upload_attachment = True
                            if record_exists:
                                existing_record = existing_records[record_id]
                                # We compare the attachment metadata and only upload if necessary.
                                existing_metadata = existing_record.get("attachment")
                                must_upload_attachment = (
                                    existing_metadata is None
                                    or existing_metadata["size"]
                                    != record_data["attachment"]["size"]
                                    or existing_metadata["hash"]
                                    != record_data["attachment"]["hash"]
                                )
                            if must_upload_attachment:
                                # If there is a .meta.json file, then read it to get original filename.
                                try:
                                    with open(f"{attachment_path}.meta.json", "rb") as f:
                                        metadata = json.load(f)
                                        filename = metadata["attachment"]["filename"]
                                except (FileNotFoundError, json.JSONDecodeError) as e:
                                    logger.error("Failed to read attachment metadata: %s", e)
                                    filename = None

                                # We upload the new attachment, and update its attributes together.
                                await async_client.add_attachment(
                                    id=record_id,
                                    bucket=bucket_id,
                                    collection=collection_id,
                                    data=record_data if load_data else None,
                                    permissions=record_permissions if load_permissions else None,
                                    filepath=attachment_path,
                                    filename=filename,
                                )

                    if not must_upload_attachment:
                        if not record_exists:
                            await batch.create_record(
                                id=record_id,
                                bucket=bucket_id,
                                collection=collection_id,
                                data=record_data if load_data else None,
                                permissions=record_permissions if load_permissions else None,
                                safe=(not force),
                            )
                        else:
                            existing_record = existing_records[record_id]
                            existing_record_data = existing_record.get("data", {})
                            existing_record_permissions = existing_record.get("permissions", {})
                            if (
                                existing_record_data != record_data
                                or existing_record_permissions != record_permissions
                            ):
                                await batch.update_record(
                                    id=record_id,
                                    bucket=bucket_id,
                                    collection=collection_id,
                                    data=record_data if load_data else None,
                                    permissions=record_permissions if load_permissions else None,
                                )

                if delete_missing_records and collection_exists and collection_records:
                    # Fetch all records IDs
                    file_records_ids = set(collection_records.keys())
                    server_records_ids = set(existing_collection["records"].keys())

                    to_delete = server_records_ids - file_records_ids
                    if not force:
                        message = (
                            "Are you sure that you want to delete the "
                            "following {} records?".format(len(list(to_delete)))
                        )
                        value = input(message)
                        if value.lower() not in ["y", "yes"]:
                            print("Exiting")
                            exit(1)
                    for record_id in to_delete:
                        await batch.delete_record(
                            id=record_id, bucket=bucket_id, collection=collection_id
                        )

        logger.debug("Sending batch:\n\n%s" % batch.session.requests)
    logger.info("Records uploaded")
